import atexit
import json
import os
import signal
import subprocess
import time
import urllib.request
from contextlib import asynccontextmanager
from enum import Enum
from pathlib import Path

import uvicorn
from connexion import AsyncApp, ConnexionMiddleware
from connexion.options import SwaggerUIOptions
from fastapi import FastAPI, Request
from filelock import SoftFileLock, Timeout
from huggingface_hub import hf_hub_download
from prefect import flow
from prefect.server.api.server import SubprocessASGIServer
from prefect.settings import (
    PREFECT_DEFAULT_WORK_POOL_NAME,
    PREFECT_SERVER_EPHEMERAL_STARTUP_TIMEOUT_SECONDS,
)
from prefect.utilities.processutils import get_sys_executable
from prefect.workers.process import ProcessWorker
from tqdm import tqdm

from timestep.api.openai.v1.controllers.completions_controller import create_completion
from timestep.config import settings
from timestep.database import create_db_and_tables

app_dir = settings.app_dir
connexion_app = AsyncApp(import_name=__name__)
default_hf_repo_id = settings.default_hf_repo_id
# default_llamafile_filename = settings.default_llamafile_filename
default_llamafile_host = settings.default_llamafile_host
default_llamafile_port = settings.default_llamafile_port
# default_llamafile_url = f"https://huggingface.co/{default_hf_repo_id}/resolve/main/{default_llamafile_filename}?download=true"
default_model_filename = settings.default_model_filename
default_multimodal_model_projector_filename = (
    settings.default_multimodal_model_projector_filename
)
# local_llamafile_path = (
#     f"{app_dir}/models/{default_hf_repo_id}/{default_llamafile_filename}"
# )
local_model_path = f"{app_dir}/models/{default_hf_repo_id}/{default_model_filename}"
local_multimodal_model_projector_path = f"{app_dir}/models/{default_hf_repo_id}/{default_multimodal_model_projector_filename}"


@asynccontextmanager
async def lifespan(app: FastAPI):
    lock_file = Path(f"{app_dir}/data/data.json")
    subprocess_manager = SubprocessManager()

    try:
        soft_lock = SoftFileLock(
            f"{str(lock_file)}.lock",
            is_singleton=True,
            thread_local=False,
            timeout=120,
        )

        worker_count = int(os.getenv("PYTEST_XDIST_WORKER_COUNT", 1))
        worker_name = os.getenv("PYTEST_XDIST_WORKER", "master")
        worker_pid = os.getpid()

        with soft_lock.acquire():
            create_db_and_tables()

            if lock_file.is_file():
                data = json.loads(lock_file.read_text())

                data["timestep"][worker_name] = {
                    "pid": worker_pid,
                    "status": "running",
                }

                lock_file.write_text(json.dumps(data))

            else:
                # create_db_and_tables()

                llamafile_pid = start_llamafile(subprocess_manager)

                prefect_server_pid = subprocess_manager.start(
                    args=[
                        "sh",
                        "-c",
                        f"PREFECT_API_URL={settings.prefect_api_url} prefect server start",
                    ],
                    name="prefect.server",
                )

                # prefect_server = SubprocessASGIServer(port=4200)
                # prefect_server.start(timeout=30)
                # prefect_server_pid = prefect_server.server_process.pid

                # print(
                #     f"Waiting {PREFECT_SERVER_EPHEMERAL_STARTUP_TIMEOUT_SECONDS.value()} seconds for Prefect server to start..."
                # )
                # time.sleep(PREFECT_SERVER_EPHEMERAL_STARTUP_TIMEOUT_SECONDS.value())
                print(
                    f"Sleeping for {PREFECT_SERVER_EPHEMERAL_STARTUP_TIMEOUT_SECONDS.value()} seconds..."
                )
                time.sleep(PREFECT_SERVER_EPHEMERAL_STARTUP_TIMEOUT_SECONDS.value())

                prefect_worker_pid = subprocess_manager.start(
                    args=[
                        "sh",
                        "-c",
                        f"PREFECT_API_URL={settings.prefect_api_url} prefect worker start --pool 'default' --type 'process'",
                    ],
                    name="prefect.worker.default",
                )

                print(f"Sleeping for 15 seconds...")
                time.sleep(15)

                data = {
                    "llamafiles": {
                        "default": {
                            "pid": llamafile_pid,
                            "status": "running",
                        }
                    },
                    "prefect": {
                        "server": {
                            "pid": prefect_server_pid,
                            "status": "running",
                        },
                        "workers": {
                            "default": {
                                "pid": prefect_worker_pid,
                                "status": "running",
                            }
                        },
                    },
                    "timestep": {
                        worker_name: {
                            "pid": worker_pid,
                            "status": "running",
                        }
                    },
                }

                lock_file.write_text(json.dumps(data))

        # TODO: Filelock doesnt support async, try https://py-filelock.readthedocs.io/en/latest/_modules/filelock/asyncio.html
        try:
            agent_flow = await flow.from_source(
                source=str(Path(__file__).parent),
                entrypoint="worker.py:agent_flow",
            )

            await agent_flow.deploy(
                name="agent-flow-deployment",
                work_pool_name="default",
            )

        except Exception as e:
            print(e)

        yield

        print(f"Exiting {worker_name} worker process with PID: {worker_pid}")

    except Exception as e:
        print(f"An error occurred: {e}; attempting to clean up...")

    finally:
        try:
            with soft_lock.acquire():
                data = json.loads(lock_file.read_text())

                print(f"Stopping {worker_name} worker process with PID: {worker_pid}")
                try:
                    data["timestep"][worker_name]["status"] = "stopped"

                except Exception as e:
                    print(e)

                stopped_workers = [
                    worker
                    for worker in data["timestep"].values()
                    if worker["status"] == "stopped"
                ]

                if len(stopped_workers) >= worker_count:
                    print("stopped_workers: ", stopped_workers)
                    print("worker_count: ", worker_count)

                    for prefect_worker_id, prefect_worker in data["prefect"][
                        "workers"
                    ].items():
                        if prefect_worker["status"] != "stopped":
                            print(
                                f"Stopping Prefect worker {prefect_worker_id} process with PID: {prefect_worker['pid']}"
                            )
                            subprocess_manager.stop(prefect_worker["pid"])

                        else:
                            print(
                                f"Prefect worker {prefect_worker_id} process with PID: {prefect_worker['pid']} already stopped"
                            )

                    if data["prefect"]["server"]["status"] != "stopped":
                        print(
                            f"Stopping Prefect server process with PID: {data['prefect']['server']['pid']}"
                        )
                        subprocess_manager.stop(data["prefect"]["server"]["pid"])

                    else:
                        print(
                            f"Prefect server process with PID: {data['prefect']['server']['pid']} already stopped"
                        )

                    for llamafile_id, llamafile in data["llamafiles"].items():
                        if llamafile["status"] != "stopped":
                            print(
                                f"Stopping llamafile {llamafile_id} process with PID: {llamafile['pid']}"
                            )
                            subprocess_manager.stop(llamafile["pid"])

                        else:
                            print(
                                f"llamafile {llamafile_id} process with PID: {llamafile['pid']} already stopped"
                            )

                    lock_file.unlink()

                else:
                    lock_file.write_text(json.dumps(data))

        except FileNotFoundError as e:
            print(f"An error occurred: {e}")

        except Exception as e:
            print(f"An error occurred in the finally block: {e}")
            raise RuntimeWarning(f"An error occurred in the finally block: {e}")


def start_llamafile(subprocess_manager):
    # if os.path.basename(
    #                 local_llamafile_path
    #             ) == default_llamafile_filename and not os.path.exists(
    #                 local_llamafile_path
    #             ):
    # os.makedirs(os.path.dirname(local_llamafile_path), exist_ok=True)
    # os.makedirs(f"{app_dir}/models/{default_hf_repo_id}", exist_ok=True)
    os.makedirs(os.path.dirname(local_model_path), exist_ok=True)
    # download_with_progress_bar(
    #                 default_llamafile_url, local_llamafile_path
    #             )
    hf_hub_download(
        # cache_dir=f"{app_dir}/models",
        # filename=default_llamafile_filename,
        filename=settings.default_model_filename,
        # local_dir=os.path.dirname(local_llamafile_path),
        local_dir=f"{app_dir}/models/{default_hf_repo_id}",
        # local_files_only=True,
        repo_id=default_hf_repo_id,
        repo_type="model",
    )

    # assert os.path.exists(local_llamafile_path)
    # assert os.path.exists(f"{app_dir}/models/{default_hf_repo_id}/{default_model_filename}")
    assert os.path.exists(local_model_path)

    if default_multimodal_model_projector_filename:
        os.makedirs(
            os.path.dirname(local_multimodal_model_projector_path), exist_ok=True
        )
        hf_hub_download(
            filename=default_multimodal_model_projector_filename,
            local_dir=f"{app_dir}/models/{default_hf_repo_id}",
            # local_files_only=True,
            repo_id=default_hf_repo_id,
            repo_type="model",
        )

        assert os.path.exists(local_multimodal_model_projector_path)

    args = ["sh"]

    if default_model_filename.endswith(".gguf"):
        args += [
            "bin/llamafile",
            "--model",
            local_model_path,
        ]

        if default_multimodal_model_projector_filename:
            args += [
                "--mmproj",
                f"{local_multimodal_model_projector_path}",
            ]

    elif default_model_filename.endswith(".llamafile"):
        # if local_model_path is not executable, make it executable
        if not os.access(local_model_path, os.X_OK):
            os.chmod(local_model_path, 0o755)

        args += [
            local_model_path,
        ]

    else:
        raise NotImplementedError(
            f"Unsupported model file type: {default_model_filename}"
        )

    args += [
        "--host",
        f"{default_llamafile_host}",
        # "--model‐draft", # TODO: Draft model for speculative decoding
        "--nobrowser",
        "--path",
        "/zip/llama.cpp/server/public",
        "--port",
        f"{default_llamafile_port}",
        "--server",
        "--temp",
        "0.0",
    ]

    # TODO: add support for LoRA adapters

    llamafile_pid = subprocess_manager.start(
        args=args,
        name="llamafiles.default",
    )

    return llamafile_pid


fastapi_app = FastAPI(
    lifespan=lifespan,
    # servers= # TODO: add server for testing?
)

connexion_app.add_api(
    "api/ap/v1/openapi/openapi.yaml",
    pythonic_params=True,
    resolver_error=501,
    swagger_ui_options=SwaggerUIOptions(serve_spec=False),
)

connexion_app.add_api(
    "api/openai/v1/openapi/openapi.yaml",
    base_path="/openai/v1",
    pythonic_params=True,
    resolver_error=500,
)

fastapi_app.mount("/api", ConnexionMiddleware(app=connexion_app, import_name=__name__))


class Engine(str, Enum):
    copilot_codex = "copilot-codex"


@fastapi_app.get("/v1/engines")
async def list_engines():
    # return [engine.value for engine in Engine]
    return {
        "data": [
            {
                "id": Engine.copilot_codex,
                "name": "Copilot Codex",
                "description": "OpenAI's Codex model, formerly known as GitHub Copilot",
            }
        ]
    }


@fastapi_app.post("/v1/engines/{engine}/completions")
async def create_code_completion(engine: Engine, request: Request):
    print(
        f"=== {__name__}.create_code_completion(engine: Engine, request: Request) ==="
    )
    print(f"engine: {engine}")
    print(f"request: {request}")

    if engine == Engine.copilot_codex:
        try:
            body = await request.json()

        except json.decoder.JSONDecodeError as e:
            print(f"Error: {e}")
            body = {
                "model": "LLaMA_CPP",
                "prompt": "int main() {",
            }

        # body["model"] = "LLaMA_CPP"

        return await create_completion(body, token_info={}, user=None)

    else:
        raise NotImplementedError("Not implemented")

    # @fastapi_app.get("/v2/models/{model_id}")
    # async def create_code_completion(model_id: str, request: Request):
    #     # model_info = models_service.retrieve_model(model_id, token_info, user)
    #     model_info = model_service.retrieve_model(model_id)

    #     print("model_info: ", model_info)

    #     return model_info


def download_with_progress_bar(
    url, filename
):  # TODO: just use from huggingface_hub import hf_hub_download and hf_hub_upload
    # Extract the basename of the file
    base_filename = os.path.basename(filename)

    # Define the reporthook function
    def reporthook(block_num, block_size, total_size):
        if pbar.total is None:
            pbar.total = total_size

        pbar.update(block_size)

    # Create a tqdm progress bar instance with a custom bar_format
    pbar = tqdm(
        unit="B",
        unit_scale=True,
        miniters=1,
        desc=base_filename,
        bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} [{percentage:.0f}%]",
    )

    # Use urlretrieve to download the file with the reporthook
    urllib.request.urlretrieve(url, filename, reporthook)

    # Close the progress bar
    pbar.close()


class SubprocessManager:
    def start(self, args, name):
        try:
            stdout_log_path = f"stdout.{name}.log"
            stderr_log_path = f"stderr.{name}.log"

            with open(stderr_log_path, "w") as stderr, open(
                stdout_log_path, "w"
            ) as stdout:
                self.process = subprocess.Popen(
                    args,
                    preexec_fn=os.setpgrp,  # Start the process in a new process group
                    stderr=stderr,
                    stdout=stdout,
                )

            print(f"Started {name} process with PID: {self.process.pid}")

            return self.process.pid

        except Exception as e:
            print(f"An error occurred: {e}")

    def stop(self, pid=None):
        if pid:
            try:
                os.killpg(os.getpgid(pid), signal.SIGTERM)

            except ProcessLookupError as e:
                print(f"An error occurred: {e}")

        else:
            print("No process to terminate")


def main(*args, **kwargs):
    print(f"=== {__name__}.main(*args, **kwargs) ===")
    print(f"args: {args}")
    print(f"kwargs: {kwargs}")

    cwd = os.getcwd()
    print(f"cwd: {cwd}")

    uvicorn.run(
        f"{__name__}:fastapi_app",
        # host="0.0.0.0",
        host=kwargs.get("host", "0.0.0.0"),
        log_level="info",
        loop="asyncio",
        # port=8000,
        port=kwargs.get("port", 8000),
        # reload=True,
        reload=kwargs.get("dev", False),
        reload_dirs=[
            f"{os.getcwd()}/timestep",
        ],
        workers=1,
    )


if __name__ == "__main__":
    main()



import os
import uuid
from abc import ABC, abstractmethod
from typing import List

from pydantic import ConfigDict
from sqlalchemy import (
    ARRAY,
    DDL,
    JSON,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    MetaData,
    Table,
    create_engine,
    event,
    func,
    text,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker
from sqlmodel import Field, SQLModel, create_engine

from timestep.config import settings

app_dir = settings.app_dir

engine = create_engine(
    # f"sqlite:///{app_dir}/data/database.db", echo=True, echo_pool=True
    f"sqlite:///{app_dir}/data/database.db"
)


class SQLModelMixin(object):
    ## TODO: Investigate inheritance and use cases of TableArtifact from Prefect
    __table_args__ = {"extend_existing": True}

    # created_at = Column(DateTime, server_default=text('NOW()'))
    created_at = Column(DateTime, default=func.now())
    updated_at = Column(DateTime, default=func.now(), onupdate=func.now())


class AgentSQLModel(SQLModel, SQLModelMixin, table=True):
    __tablename__: str = "agents"

    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    description: str | None = None
    instructions: str | None = None
    model: str = Field()
    # models: List[str] = Field(sa_column=Column(JSON), default=[])
    # model_config = ConfigDict(protected_namespaces=())
    # model_id: uuid.UUID = Field(foreign_key="models.id")
    name: str | None = None
    tools: List[dict] = Field(sa_column=Column(JSON), default=[])


class MessageSQLModel(SQLModel, SQLModelMixin, table=True):
    __tablename__: str = "messages"
    __table_args__ = {"extend_existing": True}
    # __sqlmodel_relationships__: dict = {"thread": relationship}

    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    attachments: List[dict] = Field(sa_column=Column(JSON), default=[])
    content: List[dict] = Field(sa_column=Column(JSON), default=[])
    role: str = Field()
    status: str = Field()  # status: Literal['in_progress', 'incomplete', 'completed']
    thread_id: uuid.UUID = Field(foreign_key="threads.id")
    # thread: "ThreadSQLModel" = relationship("ThreadSQLModel", back_populates="messages")

    # Relationship back to the thread
    # thread = relationship("ThreadSQLModel", back_populates="messages")


# class ModelSQLModel(SQLModel, SQLModelMixin, table=True):
#     __tablename__: str = "models"
#     id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)


# class ModelAliasSQLModel(SQLModel, SQLModelMixin, table=True):
#     __tablename__: str = "model_aliases"
#     alias: str = Field(default=None, primary_key=True)
#     model_config = ConfigDict(protected_namespaces=())
#     model_id: uuid.UUID = Field(foreign_key="models.id", ondelete="CASCADE")


class ThreadSQLModel(SQLModel, SQLModelMixin, table=True):
    __tablename__: str = "threads"
    __table_args__ = {"extend_existing": True}

    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    is_locked: bool = Field(default=False)
    # is_locked = Column(Integer, default=0)
    # is_locked: bool = Field(default=False)

    # messages: List[MessageSQLModel] = relationship("MessageSQLModel", back_populates="thread")

    # Relationship to messages
    # messages = relationship("MessageSQLModel", back_populates="thread")


class VectorSQLModel(SQLModel, SQLModelMixin, table=True):
    __tablename__: str = "vectors"

    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    # distance_to_center: float = Field()
    # metadata: dict = Field(sa_column=Column(JSON), default={})
    vector: List[float] = Field(sa_column=Column(JSON), default=[])


# DDL statements for triggers
trigger_insert_ddl = DDL(
    """
CREATE TRIGGER prevent_insert_on_locked_threads
BEFORE INSERT ON messages
FOR EACH ROW
WHEN EXISTS (
    SELECT 1
    FROM threads
    WHERE threads.id = NEW.thread_id AND threads.is_locked = 1
)
BEGIN
    SELECT RAISE(ABORT, 'Operation not allowed: the thread is locked');
END;
"""
)

trigger_update_ddl = DDL(
    """
CREATE TRIGGER prevent_update_on_locked_threads
BEFORE UPDATE ON messages
FOR EACH ROW
WHEN EXISTS (
    SELECT 1
    FROM threads
    WHERE threads.id = NEW.thread_id AND threads.is_locked = 1
)
BEGIN
    SELECT RAISE(ABORT, 'Operation not allowed: the thread is locked');
END;
"""
)

trigger_delete_ddl = DDL(
    """
CREATE TRIGGER prevent_delete_on_locked_threads
BEFORE DELETE ON messages
FOR EACH ROW
WHEN EXISTS (
    SELECT 1
    FROM threads
    WHERE threads.id = OLD.thread_id AND threads.is_locked = 1
)
BEGIN
    SELECT RAISE(ABORT, 'Operation not allowed: the thread is locked');
END;
"""
)

# Attach triggers to the table creation event
event.listen(MessageSQLModel.__table__, "after_create", trigger_insert_ddl)
event.listen(MessageSQLModel.__table__, "after_create", trigger_update_ddl)
event.listen(MessageSQLModel.__table__, "after_create", trigger_delete_ddl)


def create_db_and_tables():
    # try:
    SQLModel.metadata.create_all(
        bind=engine,
        # checkfirst=False,
        checkfirst=True,
        tables=None,
        # tables=[
        #     AgentSQLModel,
        #     # ModelSQLModel,
        #     # ModelAliasSQLModel,
        #     MessageSQLModel,
        #     ThreadSQLModel,
        # ],
    )

    # except Exception as e:
    #     print(f"Error creating tables: {e}")


if __name__ == "__main__":
    create_db_and_tables()



import asyncio
import os

from openai import AsyncOpenAI, AsyncStream
from openai.types.create_embedding_response import CreateEmbeddingResponse
from sse_starlette import EventSourceResponse

from timestep.config import settings


async def create_embedding(body, token_info: dict, user: str, **kwargs):
    """Creates an embedding vector representing the input text.

    :param create_embedding_request:
    :type create_embedding_request: dict | bytes

    :rtype: Union[CreateEmbeddingResponse, Tuple[CreateEmbeddingResponse, int], Tuple[CreateEmbeddingResponse, int, Dict[str, str]]
    """
    print("=== create_embedding ===")
    # print('args:', args)
    print("body:", body)
    print("kwargs:", kwargs)

    client = AsyncOpenAI(
        api_key=settings.openai_api_key.get_secret_value(),
        base_url="http://localhost:8080/v1",
    )

    # completion: Completion | AsyncStream[Completion] = await client.completions.create(
    create_embedding_response: (
        CreateEmbeddingResponse | AsyncStream[CreateEmbeddingResponse]
    ) = await client.embeddings.create(
        **body,
        # user=user,
    )

    if type(create_embedding_response) == AsyncStream:

        async def event_publisher():
            try:
                async for chunk in create_embedding_response:
                    yield chunk.model_dump_json()

            except asyncio.CancelledError as e:
                print(f"Disconnected from client (via refresh/close)")
                # Do any other cleanup, if any
                raise e

        return EventSourceResponse(event_publisher())

    else:
        return create_embedding_response.model_dump(mode="json")



import os

import typer
from pydantic import Field, SecretStr
from pydantic_settings import BaseSettings

app_dir = typer.get_app_dir(__package__)

os.makedirs(f"{app_dir}/data", exist_ok=True)
os.makedirs(f"{app_dir}/models", exist_ok=True)
os.makedirs(f"{app_dir}/secrets", exist_ok=True)


class Settings(BaseSettings):
    app_dir: str = Field(default=app_dir)
    bearerinfo_func: str = Field(default="timestep.api.decode_token")
    default_hf_repo_id: str = Field(
        default="Mozilla/TinyLlama-1.1B-Chat-v1.0-llamafile"
        # default="squeeze-ai-lab/TinyAgent-1.1B-GGUF",
        # default="RichardErkhov/upstage_-_TinySolar-248m-4k-py-gguf",
        # default="mjschock/SmolLM-135M-Q4_K_M-GGUF",
        # default="mjschock/TinySolar-248m-4k-code-instruct-Q4_K_M-GGUF",
        # default="mjschock/TinySolar-248m-4k-py-Q4_K_M-GGUF",
        # default="mjschock/TinyLlama-1.1B-Chat-v1.0-Q4_K_M-GGUF",
        # default="mjschock/TinyLlama-1.1B-Chat-v1.0-Tinygrad"
        # default="TinyLlama/TinyLlama-1.1B-Chat-v1.0",
    )
    # default_hf_repo_id: str = Field(default="jartine/TinyLlama-1.1B-Chat-v1.0-GGUF")
    # default_hf_repo_id: str = Field(default="jartine/rocket-3B-llamafile")
    # default_hf_repo_id: str = Field(default="jartine/phi-2-llamafile")
    # default_llamafile_filename: str = Field(
    #     default="TinyLlama-1.1B-Chat-v1.0.F16.llamafile"
    # )
    # default_llamafile_filename: str = Field(default="TinyLlama-1.1B-Chat-v1.0.Q5_K_M.llamafile")
    # default_llamafile_filename: str = Field(default="rocket-3b.Q5_K_M.llamafile")
    # default_llamafile_filename: str = Field(default="phi-2.Q5_K_M.llamafile")
    default_llamafile_host: str = Field(default="0.0.0.0")
    default_llamafile_port: int = Field(default=8080)
    default_model_filename: str = Field(
        default="TinyLlama-1.1B-Chat-v1.0.F16.llamafile"
        # default="TinyAgent-1.1B-Q4_K_M.gguf",
        # default="TinyAgent-1.1B-F16.gguf",
        # default="TinySolar-248m-4k-py.Q8_0.gguf",
        # default="smollm-135m-q4_k_m.gguf", # TODO: what's a imatrix.data file and should it be used?
        # default="tinysolar-248m-4k-code-instruct-q4_k_m.gguf",
        # default="tinysolar-248m-4k-py-q4_k_m.gguf",
        # default="tinyllama-1.1b-chat-v1.0-q4_k_m.gguf",
        # default="Models-1.1B-F16.gguf",
        # default="model.safetensors",
    )
    default_multimodal_model_projector_filename: str | None = Field(
        default=None,
    )
    openai_api_key: SecretStr = Field(default="openai_api_key")
    openai_base_url: str = Field(default="http://localhost:8000/api/openai/v1")
    openai_org_id: str = Field(default="organization_id")
    openai_project_id: str = Field(default="project_id")
    poetry_repositories_testpypi_url: str = Field(
        default="https://test.pypi.org/legacy/"
    )
    poetry_virtualenvs_in_project: bool = Field(default=True)
    poetry_virtualenvs_prefer_active_python: bool = Field(default=True)
    prefect_api_url: str = Field(default="http://localhost:4200/api")
    prefect_logging_level: str = Field(default="INFO")
    prefect_logging_log_prints: bool = Field(default=True)
    pyenv_version: str = Field(default="3.10.14")

    # poetry_pypi_token_testpypi: SecretStr = Field(default=None)
    # poetry_virtualenvs_create: bool = Field()
    # poetry_virtualenvs_in_project: bool = Field()
    # prefect_api_key: Optional[SecretStr] = Field(default=None)
    # prefect_api_url: str = Field(
    #     default="http://prefect-server.default.svc.cluster.local:4200/api"
    # )
    # pyenv_version: str = Field()
    verbose: bool = Field(default=True)
    # version: str = Field()

    class Config:
        env_file = ".env"
        env_file_encoding = "utf-8"
        # secrets_dir = "./secrets"  # TODO: Change to f"{app_dir}/secrets" when ready
        secrets_dir = f"{app_dir}/secrets"


settings = Settings()

if "OPENAI_API_KEY" not in os.environ:
    os.environ["OPENAI_API_KEY"] = settings.openai_api_key.get_secret_value()

if "OPENAI_BASE_URL" not in os.environ:
    os.environ["OPENAI_BASE_URL"] = settings.openai_base_url

if "PREFECT_API_URL" not in os.environ:
    os.environ["PREFECT_API_URL"] = settings.prefect_api_url



import os
import time
from typing import List, Optional

import controlflow as cf
import httpx
import openai
import typer
from langchain_openai import ChatOpenAI
from openai.pagination import AsyncCursorPage, SyncCursorPage
from openai.types.beta.assistant import Assistant
from openai.types.beta.thread import Thread
from openai.types.beta.threads.image_file_content_block import ImageFileContentBlock
from openai.types.beta.threads.image_url_content_block import ImageURLContentBlock
from openai.types.beta.threads.message import Attachment, Message, MessageContent
from openai.types.beta.threads.run import AssistantTool, Run
from openai.types.beta.threads.runs import (
    MessageCreationStepDetails,
    RunStep,
    ToolCallsStepDetails,
)
from openai.types.beta.threads.runs.message_creation_step_details import MessageCreation
from openai.types.beta.threads.runs.tool_calls_step_details import (
    ToolCall,
    ToolCallsStepDetails,
)
from openai.types.beta.threads.text_content_block import TextContentBlock
from openai.types.chat.chat_completion import ChatCompletion, Choice
from openai.types.chat.chat_completion_assistant_message_param import (
    ChatCompletionAssistantMessageParam,
)
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai.types.chat.chat_completion_function_message_param import (
    ChatCompletionFunctionMessageParam,
)
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
)
from openai.types.chat.chat_completion_stream_options_param import (
    ChatCompletionStreamOptionsParam,
)
from openai.types.chat.chat_completion_system_message_param import (
    ChatCompletionSystemMessageParam,
)
from openai.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionToolChoiceOptionParam,
)
from openai.types.chat.chat_completion_tool_message_param import (
    ChatCompletionToolMessageParam,
)
from openai.types.chat.chat_completion_tool_param import ChatCompletionToolParam
from openai.types.chat.chat_completion_user_message_param import (
    ChatCompletionUserMessageParam,
)
from openai.types.file_object import FileObject
from openai.types.fine_tuning.fine_tuning_job import FineTuningJob, Hyperparameters
from openai.types.shared_params.function_definition import FunctionDefinition
from openai.types.shared_params.function_parameters import FunctionParameters
from prefect import flow, get_run_logger, task
from prefect.context import get_run_context
from prefect.flow_runs import pause_flow_run
from prefect.variables import Variable
from prefect_shell import ShellOperation
from pydantic import BaseModel

from timestep.config import settings

app_dir = typer.get_app_dir("timestep")
cf.default_model = ChatOpenAI(
    api_key=settings.openai_api_key.get_secret_value(),
    base_url=settings.openai_base_url,
    temperature=0.0,
)
openai_async_client = openai.AsyncOpenAI(
    api_key=settings.openai_api_key.get_secret_value(),
    base_url=settings.openai_base_url,
)


class RunInput(BaseModel):
    agent_id: str
    instructions: str
    thread_id: str
    tools: List[AssistantTool] = []


class RunOutput(BaseModel):
    steps: List[RunStep] = []


# class RunStepInput(RunInput):  # TODO: StepInput?
#     pass
#     # agent_id: str
#     # instructions: str = ""
#     # thread_id: str
#     # tools: List[AssistantTool] = []


# class RunStepOutput(BaseModel):  # TODO: StepOutput?
#     chat_completion: ChatCompletion
#     finish_reason: str
#     step_details: dict


async def lock_thread(flow, flow_run, state):
    logger = get_run_logger()
    logger.info("lock_thread")

    thread_id = get_run_context().parameters["thread_id"]
    logger.info(f"thread_id: {thread_id}")

    # lock the thread_id
    await Variable.set(f"lock_{thread_id}", "locked")

    raise NotImplementedError("Need to implement a lock mechanism")


async def unlock_thread(flow, flow_run, state):
    logger = get_run_logger()
    logger.info("unlock_thread")

    thread_id = get_run_context().parameters["thread_id"]
    logger.info(f"thread_id: {thread_id}")

    # unlock the thread_id
    await Variable.set(f"lock_{thread_id}", "unlocked")

    raise NotImplementedError("Need to implement an unlock mechanism")


@flow(
    log_prints=True,
    on_completion=[unlock_thread],
    on_running=[lock_thread],
)
# @cf.flow(timeout_seconds=300)
async def agent_flow(run_input: RunInput) -> RunOutput:
    # See https://platform.openai.com/docs/assistants/deep-dive/managing-threads-and-messages
    # for more information on managing threads and messages

    logger = get_run_logger()
    logger.info(f"run_input: {run_input}")

    agent_id = run_input.agent_id
    logger.info(f"agent_id: {agent_id}")

    thread_id = run_input.thread_id  # TODO: introduce a lock on the thread_id
    logger.info(f"thread_id: {thread_id}")

    assistant: Assistant = await openai_async_client.beta.assistants.retrieve(
        assistant_id=agent_id
    )
    logger.info(f"assistant: {assistant}")

    # flow_run_id = cf.context.get("flow_run_id")
    flow_run_id = get_run_context().flow_run.id
    run_id = str(flow_run_id).replace("-", "")

    run_output = RunOutput(
        **await Variable.get(
            f"run_{run_id}", default=RunOutput().model_dump(mode="json")
        )
    )

    step_count = 0

    while step_count < 10:
        # run_step_input = RunStepInput(
        #     agent_id=agent_id,
        #     instructions=run_input.instructions,
        #     thread_id=thread_id,
        #     tools=run_input.tools,
        # )

        # TODO: optionally, allow the agent to learn from this thread
        # run_step_output: RunStepOutput = await agent_step_task(run_step_input)
        # run_step: RunStep = await agent_step_task(run_input)

        async_cursor_page: AsyncCursorPage[Message] = (
            await openai_async_client.beta.threads.messages.list(
                order="asc", thread_id=thread_id
            )
        )
        logger.info(f"async_cursor_page: {async_cursor_page}")

        messages: List[Message] = async_cursor_page.data
        logger.info(f"messages: {messages}")

        chat_completion_message_params: List[ChatCompletionMessageParam] = []

        for message in messages:
            content: List[MessageContent] = message.content

            for _content in content:
                if type(_content) == TextContentBlock:
                    chat_completion_message_params.append(
                        ChatCompletionUserMessageParam(
                            content=_content.text.value,
                            role=message.role,
                        )
                    )

                else:
                    logger.info(f"_content: {_content}")

                    raise NotImplementedError(
                        f"Unsupported content type: {type(_content)}"
                    )

        logger.info(f"chat_completion_message_params: {chat_completion_message_params}")

        chat_completion_tool_params = build_chat_completion_tool_params(
            run_input, assistant
        )

        logger.info(f"chat_completion_tool_params: {chat_completion_tool_params}")

        chat_completion: ChatCompletion = (
            await openai_async_client.chat.completions.create(
                messages=chat_completion_message_params,
                model=assistant.model,
                tools=chat_completion_tool_params,
            )
        )

        logger.info(f"chat_completion: {chat_completion}")

        choices: List[Choice] = chat_completion.choices
        logger.info(f"choices: {choices}")

        if len(choices) > 1:
            raise NotImplementedError("Multiple choices not yet supported")

        first_choice: Choice = choices[
            0
        ]  # TODO: check out Choice.logprobs for picking the best choice
        logger.info(f"first_choice: {first_choice}")

        # finish_reason: Literal["stop", "length", "tool_calls", "content_filter", "function_call"]
        # """The reason the model stopped generating tokens.

        # This will be `stop` if the model hit a natural stop point or a provided stop
        # sequence, `length` if the maximum number of tokens specified in the request was
        # reached, `content_filter` if content was omitted due to a flag from our content
        # filters, `tool_calls` if the model called a tool, or `function_call`
        # (deprecated) if the model called a function.
        # """
        finish_reason: str = first_choice.finish_reason
        logger.info(f"finish_reason: {finish_reason}")

        chat_completion_message: ChatCompletionMessage = first_choice.message
        logger.info(f"chat_completion_message: {chat_completion_message}")

        tool_calls: List[ChatCompletionMessageToolCall] | None = (
            chat_completion_message.tool_calls
        )
        logger.info(f"tool_calls: {tool_calls}")

        # Step Details: https://platform.openai.com/docs/api-reference/run-steps/step-object#run-steps/step-object-step_details
        if tool_calls:
            # Set the run status to "requires_action" and "required_action.type" to "submit_tool_outputs"
            # Pause this flow run until the required action is completed (e.g. label = await pause_flow_run(wait_for_input=Animal, timeout=60 * 60))
            step_details = ToolCallsStepDetails(
                tool_calls=tool_calls,
                type="tool_calls",
            )

            raise NotImplementedError

        else:
            # Message Creation Run Step
            message: Message = await openai_async_client.beta.threads.messages.create(
                thread_id=thread_id,
                # content: str | Iterable[MessageContentPartParam],
                content=chat_completion_message.content,
                # role: Literal['user', 'assistant'],
                role=chat_completion_message.role,
                # run_id=run_id, # TODO: add run_id to the message created here
                # attachments: Iterable[Attachment] | NotGiven | None = NOT_GIVEN,
                # metadata: object | NotGiven | None = NOT_GIVEN,
                # extra_headers: Headers | None = None,
                # extra_query: Query | None = None,
                # extra_body: Body | None = None,
                # timeout: float | Timeout | NotGiven | None = NOT_GIVEN
            )

            step_details = MessageCreationStepDetails(
                message_creation=MessageCreation(
                    message_id=message.id,
                ),
                type="message_creation",
            )

        # run_step_output = RunStepOutput(
        #     chat_completion=chat_completion,
        #     finish_reason=finish_reason,
        #     step_details=step_details,
        # )

        # return run_step_output

        run_step = RunStep(
            id=f"step-{step_count}",
            created_at=int(time.time()),
            assistant_id=agent_id,
            object="thread.run.step",
            run_id=run_id,
            status="completed",
            step_details=step_details,
            thread_id=thread_id,
            type=step_details.type,
        )

        # run_output["steps"].append(run_step_output)
        # run_output.steps.append(run_step_output)
        run_output.steps.append(run_step)
        await Variable.set(f"run_{run_id}", run_output.model_dump(mode="json"))
        # await Variable.set("run_output", run_output)

        step_count += 1

        # if run_step_output.finish_reason == "stop":
        # if True:
        if finish_reason == "stop":
            break

        else:
            raise NotImplementedError(
                # f"Unsupported finish_reason: {run_step_output.finish_reason}"
                f"Unsupported finish_reason: {finish_reason}"
            )

    return run_output


def build_chat_completion_tool_params(run_input, assistant):
    chat_completion_tool_params: List[ChatCompletionToolParam] = []
    # chat_completion_tool_params: [
    #     {'function': {'description': 'code_interpreter', 'name': 'code_interpreter', 'parameters': {}}, 'type': 'function'},
    #     {'function': {'description': 'file_search', 'name': 'file_search', 'parameters': {}}, 'type': 'function'},
    #     {'function': {'description': 'function', 'name': 'function', 'parameters': {}}, 'type': 'function'}]
    # See https://platform.openai.com/docs/assistants/tools/code-interpreter; TODO: use default input: str, outputs: list, etc for code_interpreter tool
    # See https://platform.openai.com/docs/assistants/tools/file-search: TODO index the files as documents if not already done using pypdf plust lightweight OCR model for blank pages, get good set of default function signature (like above)

    # Function tool example:
    #   "name": "get_current_weather",
    #   "description": "Get the current weather in a given location",
    #   "parameters": {
    #     "type": "object",
    #     "properties": {
    #       "location": {
    #         "type": "string",
    #         "description": "The city and state, e.g. San Francisco, CA",
    #       },
    #       "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
    #     },
    #     "required": ["location"],
    #   },

    # for tool in run_step_input.tools or assistant.tools:
    tools = run_input.tools or assistant.tools

    print("run_input.tools: ", run_input.tools)

    for tool in run_input.tools:
        print("tool: ", tool)
        print("type(tool): ", type(tool))

    print("assistant.tools: ", assistant.tools)

    for tool in assistant.tools:
        print("tool: ", tool)
        print("type(tool): ", type(tool))

    for tool in tools:
        assert tool.type in [
            "code_interpreter",
            "file_search",
            "function",
        ], f"Unsupported tool type: {tool.type}"

        function_description: str = tool.type
        function_name: str = tool.type
        function_parameters: FunctionParameters = dict()

        chat_completion_tool_params.append(
            ChatCompletionToolParam(
                function=FunctionDefinition(
                    description=function_description,
                    name=function_name,
                    parameters=function_parameters,
                ),
                type="function",
            )
        )

    return chat_completion_tool_params


# @task
# # @cf.task(timeout_seconds=60)
# # async def agent_step_task(run_step_input: RunStepInput):
# # async def agent_step_task(run_step_input: Run) -> RunStep:
# async def agent_step_task(run_step_input: RunInput) -> RunStep:
#     logger = get_run_logger()
#     logger.info(f"run_step_input: {run_step_input}")

#     agent_id = run_step_input.agent_id
#     logger.info(f"agent_id: {agent_id}")

#     thread_id = run_step_input.thread_id
#     logger.info(f"thread_id: {thread_id}")


# @task
# async def select_fine_tuning_job_by_id(id: str):
#     async with await SqlAlchemyConnector.load(block_name) as connector:
#         # results = await connector.fetch_one(f"SELECT * FROM fine_tuning_jobs WHERE id = '{unhyphenated_id}';") # TODO: use prepared statement equivalent
#         results = await connector.fetch_one(
#             f"""SELECT
#                     created_at,
#                     error,
#                     estimated_finish,
#                     fine_tuned_model,
#                     finished_at,
#                     hyperparameters,
#                     integrations,
#                     model,
#                     object,
#                     organization_id,
#                     result_files,
#                     seed,
#                     status,
#                     trained_tokens,
#                     training_file,
#                     validation_file
#                 FROM fine_tuning_jobs
#                 WHERE id = :id;""",
#             parameters={"id": str(uuid.UUID(id, version=4)).replace("-", "")},
#         )
#         print("results: ", results)

#         return FineTuningJob(
#             id=id,
#             created_at=results[0],
#             error=json.loads(results[1]),
#             estimated_finish=results[2],
#             fine_tuned_model=results[3],
#             finished_at=results[4],
#             hyperparameters=json.loads(results[5]),
#             integrations=json.loads(results[6]),
#             model=results[7],
#             object=results[8],
#             organization_id=results[9],
#             result_files=json.loads(results[10]),
#             seed=results[11],
#             status=results[12],
#             trained_tokens=results[13],
#             training_file=results[14],
#             validation_file=results[15],
#         )
#         # return results

#     # return all_rows


@flow(log_prints=True)
async def agent_train_step_flow(fine_tuning_job_id: str, suffix: Optional[str] = None):
    logger = get_run_logger()
    logger.info("INFO level log message.")

    print("=== training model ===")
    print("fine_tuning_job_id: ", fine_tuning_job_id)

    working_dir = f"{app_dir}/work/{fine_tuning_job_id}"
    os.makedirs(working_dir, exist_ok=True)

    # fine_tuning_job: FineTuningJob = instance_store._shared_instance_state["fine_tuning_jobs"].get(fine_tuning_job_id)
    # fine_tuning_job: FineTuningJob = await select_fine_tuning_job_by_id(
    #     fine_tuning_job_id
    # )

    openai_client = openai.OpenAI(
        api_key=os.environ.get("OPENAI_API_KEY"),
        base_url=os.environ.get("OPENAI_BASE_URL"),
    )

    fine_tuning_job: FineTuningJob = (
        openai_client.fine_tuning.jobs.retrieve(  # TODO: is there an async version?
            fine_tuning_job_id=fine_tuning_job_id,
        )
    )

    print("fine_tuning_job: ", fine_tuning_job)

    hyperparameters: Hyperparameters = fine_tuning_job.hyperparameters
    max_epochs = -1 if hyperparameters.n_epochs == "auto" else hyperparameters.n_epochs

    # model: Llama = instance_store._shared_instance_state["models"][fine_tuning_job.model]
    async with httpx.AsyncClient() as client:
        response = await client.get(
            f'{os.environ.get("OPENAI_BASE_URL").split("/api")[0]}/v2/models/{fine_tuning_job.model}'
        )
        print("response: ", response)
        model_info = response.json()
        print("model_info: ", model_info)

    model_path: str = model_info.get("model_path")
    print("model_path: ", model_path)

    lora_base: str = model_info.get("lora_base")
    print("lora_base: ", lora_base)

    lora_path: str = model_info.get("lora_path")
    print("lora_path: ", lora_path)

    lora_scale: str = model_info.get("lora_scale")
    print("lora_scale: ", lora_scale)

    # fine_tuning_job = FineTuningJob(
    #     id=str(uuid.uuid4()),
    #     created_at=int(time.time()),
    #     hyperparameters=hyperparameters,
    #     model=body.get("model"),
    #     object="fine_tuning.job",
    #     organization_id="organization_id",
    #     result_files=[],
    #     seed=body.get("seed", 42),
    #     status="queued", # Literal['validating_files', 'queued', 'running', 'succeeded', 'failed', 'cancelled'],
    #     training_file=body.get("training_file"),
    #     validation_file=body.get("validation_file"),
    # )

    # fine_tuning_jobs_timestep.db[fine_tuning_job.id] = fine_tuning_job

    # fine_tuning_job.__setattr__(
    #     "status",
    #     "running",
    # )

    # fine_tuning_job_event = FineTuningJobEvent(
    #     id=str(uuid.uuid4()),
    #     created_at=int(time.time()),
    #     level="info",
    #     message=f"The job has begun",
    #     object="fine_tuning.job.event",
    # )

    # # train_data = f"{os.getcwd()}/notebooks/openai-cookbook/examples/tmp_recipe_finetune_training.jsonl"
    # # train_data = fine_tuning_job.training_file
    training_file_id = fine_tuning_job.training_file
    validation_file_id = fine_tuning_job.validation_file

    # print('training_file_id: ', training_file_id)
    # print('validation_file_id: ', validation_file_id)

    # training_file_object = instance_store._shared_instance_state["file_objects"][training_file_id]
    # validation_file_object = instance_store._shared_instance_state["file_objects"][validation_file_id]

    training_file_object: FileObject = openai_client.files.retrieve(
        file_id=training_file_id
    )

    print("training_file_object: ", training_file_object)

    if validation_file_id:
        validation_file_object: FileObject = openai_client.files.retrieve(
            file_id=validation_file_id
        )

        print("validation_file_object: ", validation_file_object)

    # str(uuid.UUID(id, version=4)).replace("-", "")

    # for long running operations, you can use a context manager
    async with ShellOperation(
        commands=[
            # TODO: https://github.com/ggerganov/llama.cpp/blob/master/docs/backend/SYCL.md
            f"git clone https://github.com/ggerganov/llama.cpp {app_dir}/3rdparty/llama.cpp",
            f"make -C {app_dir}/3rdparty/llama.cpp",  # TODO: handle windows, termux, etc.
            f"""{app_dir}/3rdparty/llama.cpp/llama-finetune \
--adam-iter 30 \
--batch 8 \
--checkpoint-in "chk-lora-{fine_tuning_job.model}-{fine_tuning_job.id}-LATEST.gguf" \
--checkpoint-out "chk-lora-{fine_tuning_job.model}-{fine_tuning_job.id}-ITERATION.gguf" \
--ctx 64 \
--epochs {max_epochs} \
--lora-out "lora-{fine_tuning_job.model}-{fine_tuning_job.id}-ITERATION.bin"  \
--model-base {model_path} \
--n-gpu-layers 0 \
--save-every 10 \
--seed {fine_tuning_job.seed} \
--threads 6 \
--train-data "{app_dir}/data/{training_file_object.id}/{training_file_object.filename}" \
--use-checkpointing
            """,
        ],
        env={
            "CMAKE_ARGS": os.getenv("CMAKE_ARGS", ""),
        },
        working_dir=working_dir,
    ) as llama_finetune_operation:

        # trigger runs the process in the background
        llama_finetune_process = await llama_finetune_operation.trigger()

        # then do other things here in the meantime, like download another file
        ...

        # when you're ready, wait for the process to finish
        await llama_finetune_process.wait_for_completion()

        # if you'd like to get the output lines, you can use the `fetch_result` method
        output_lines = await llama_finetune_process.fetch_result()
        print("output_lines: ", output_lines)

    #     for output_line in output_lines:
    #         print('output_line: ', output_line)

    #         fine_tuning_job_event = FineTuningJobEvent(
    #             id=str(uuid.uuid4()),
    #             created_at=int(time.time()),
    #             level="info",
    #             message=output_line,
    #             object="fine_tuning.job.event",
    #         )

    #         instance_store._shared_instance_state["fine_tuning_job_events"][fine_tuning_job.id].append(fine_tuning_job_event)

    #     fine_tuning_job.__setattr__(
    #         "fine_tuned_model",
    #         # "ft:gpt-3.5-turbo-0613:personal:recipe-ner:8PjmcwDH"
    #         f"ft:{fine_tuning_job.model}:{fine_tuning_job.organization_id}:{suffix}:{fine_tuning_job.id}"
    #     )

    #     fine_tuning_job_event = FineTuningJobEvent(
    #         id=str(uuid.uuid4()),
    #         created_at=int(time.time()),
    #         level="info",
    #         message=f"New fine-tuned model created: {fine_tuning_job.fine_tuned_model}",
    #         object="fine_tuning.job.event",
    #     )

    #     instance_store._shared_instance_state["fine_tuning_job_events"][fine_tuning_job.id].append(fine_tuning_job_event)

    #     fine_tuning_job.__setattr__(
    #         "status",
    #         "succeeded",
    #     )

    #     fine_tuning_job_event = FineTuningJobEvent(
    #         id=str(uuid.uuid4()),
    #         created_at=int(time.time()),
    #         level="info",
    #         message="The job has successfully completed",
    #         object="fine_tuning.job.event",
    #     )

    #     instance_store._shared_instance_state["fine_tuning_job_events"][fine_tuning_job.id].append(fine_tuning_job_event)

    #     base_model: Llama = instance_store._shared_instance_state["models"][fine_tuning_job.model]

    #     # model = Llama(
    #     #     chat_format=base_model.chat_format,
    #     #     lora_base=f"{os.getcwd()}/{base_model.model_path}",
    #     #     lora_path=f"{os.getcwd()}/{working_dir}/lora-{fine_tuning_job.model}-{task}-ITERATION.bin",
    #     #     model_path=base_model.model_path,
    #     #     n_ctx=base_model.n_ctx(),
    #     # )

    #     # model = serve_model()

    #     # instance_store._shared_instance_state["models"][fine_tuning_job.fine_tuned_model] = model


# async def main():
# def main(work_type="train_agent"):
def main(thread_id=None):
    raise NotImplementedError

    # flow_run: FlowRun = await run_deployment(
    #     idempotency_key=thread_id,
    #     name="agent-flow/agent-flow-deployment",
    #     parameters={
    #         "inputs": {
    #             "agent_id": assistant_id,
    #             "thread_id": thread_id,
    #         }
    #     },
    #     # job_variables={"env": {"MY_ENV_VAR": "staging"}},
    #     timeout=0,  # don't wait for the run to finish
    # )


if __name__ == "__main__":
    # main(work_type="train_agent")
    main()
